import os
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from sklearn.datasets import make_moons, make_blobs
from StructuredGraphLearning.LearnGraphTopology import LearnGraphTopology

if __name__ == "__main__":
    plots_dir = './plots'
    if not os.path.exists(plots_dir):
        os.makedirs(plots_dir)

    np.random.seed(0)
    n = 50  # number of nodes per cluster
    k = 2   # number of components
    X, y = make_moons(n_samples=n*k, noise=.05, shuffle=True)
    # X, y = make_blobs(n_samples=n*k, centers=k, n_features=2, random_state=0)
    # dict to store position of nodes
    pos = {}
    for i in range(n*k):
        pos[i] = X[i]
    # Visualization of original data
    fig = plt.figure()
    plt.scatter(X[:,0], X[:,1], c=y )
    plt.title("Two moon dataset")
    plt.xlabel('x-coordinate')
    plt.ylabel('y-coordinate')
    # fig.savefig('plots/two_moon_dataset.eps', format='eps')
    # fig.savefig('plots/two_moon_dataset.png')

    # compute sample correlation matrix
    S = np.dot(X, X.T)

    # estimate underlying graph
    sgl = LearnGraphTopology(S, maxiter=1000, record_objective = True, record_weights = True)
    graph = sgl.learn_k_component_graph(k=2, beta=0.1 )

    nll = graph['nll']
    print('NLL: ', min(nll))
    objective = graph['obj_fun']
    print('Objective: ', min(objective))

    # build network
    A = graph['adjacency']
    G = nx.from_numpy_matrix(A)
    print('Graph statistics:')
    print('Nodes: ', G.number_of_nodes(), 'Edges: ', G.number_of_edges() )

    # normalize edge weights to plot edges strength
    all_weights = []
    for (node1,node2,data) in G.edges(data=True):
        all_weights.append(data['weight'])
    max_weight = max(all_weights)
    norm_weights = [3* w / max_weight for w in all_weights]
    norm_weights = norm_weights

    # plot graph
    fig = plt.figure(figsize=(15,15))
    nx.draw_networkx(G,pos, width=norm_weights)
    plt.title("Learned graph for two moon dataset")
    plt.suptitle('components k=2')
    plt.xlabel('x-coordinate')
    plt.ylabel('y-coordinate')
    # filename = 'plots/learned_graph_k='+ str(k) +'.eps'
    # png_filename = 'plots/learned_graph_k='+ str(k) +'.png'
    # fig.savefig(filename, format='eps')
    # fig.savefig(png_filename,)


import numpy as np

class Metrics:
    def __init__(self, Lw_true, Lw_est, eps=1e-4):
        self.Lw_true = Lw_true
        self.Lw_est = Lw_est
        self.eps = eps
        n = self.Lw_true.shape[0]
        self.tp, self.fp, self.fn, self.tn = 0, 0, 0, 0
        for i in range(0, n-1):
            for j in range(i+1, n):
                is_edge = abs(self.Lw_true[i][j]) > eps
                is_est_edge = abs(self.Lw_est[i][j]) > eps
                if is_edge and is_est_edge:
                    self.tp += 1
                elif not is_edge and is_est_edge:
                    self.fp += 1
                elif is_edge and not is_est_edge:
                    self.fn += 1
                else:
                    self.tn += 1

    def relative_error(self):
        """Relative Error between the final estimated laplacian and the true reference
        graph Laplacian matrix. 

        Parameters
        ----------
        Lw_true : true Laplacian
        Lw_est : estimated laplacian
        """
        return np.linalg.norm(self.Lw_true-self.Lw_est, ord=2) / np.linalg.norm(self.Lw_true, ord=2)
        
    def f1_score(self):
        return 2 * self.tp / (2 * self.tp + self.fn + self.fp)
    
    def recall(self):
        return self.tp / (self.tp + self.fn)
    
    def specificity(self):
        self.tn / (self.tn + self.fp)

    def accuracy(self):
        (self.tp + self.tn) / (self.tp + self.tn + self.fp + self.fn)

    def negative_predictive_value(self):
        return (self.tn / (self.tn + self.fn))

    def false_discovery_rate(self):
        return (self.fp / (self.fp + self.tp))

class ModelSelection:
    def gdet(self, L):
        eps = 1e-4
        vals = np.linalg.eigvals(L)
        nz_eigvals = vals[vals > eps]
        return np.prod(nz_eigvals)

    def bic(self, L, S, n, p):
        """Bayesian information criterion
        eBIC(\theta) = log gdet(\theta) − Tr(S\theta) − |E|log n
        """
        A = np.diag(np.diag(L)) - L 
        num_edges = np.sum(A[A>0])/2
        return 2*np.log(self.gdet(L)) - np.trace(S*L) - num_edges * np.log(n)

    def ebic(self, L, S, n, p):
        """Extended Bayesian information criterion
        eBIC(\theta) = log gdet(\theta) − Tr(S\theta) − |E|log n − 4γ|E|log p
        """
        # if n/p > 500:
        #     gamma = 0
        # elif n/p > 100:
        #     gamma = 0.5
        # else:
        gamma = 0
        A = np.diag(np.diag(L)) - L 
        num_edges = np.sum(A[A>0])/2
        ll = np.log(self.gdet(L)) - np.trace(S @ L)
        print(2*ll, num_edges * np.log(n), num_edges)
        return 2*ll - num_edges * np.log(n) - 4*gamma*num_edges*np.log(p)

    # def aic(self, L, S, n, p):
    #     """Akaike information criterion
    #     eBIC(\theta) = log gdet(\theta) − Tr(S\theta) − |E|log n − 4γ|E|log p
    #     """
    #     return -2 * ll + 2 * k

    # def aicc(self, ll, n, k):
    #     """Akaike information criterion corrected

    #     """
    #     return self.aic(ll,n, k) + 2*k*(k+1)/(n-k-1)


import numpy as np
import math
import time
from StructuredGraphLearning.utils import Operators
from StructuredGraphLearning.optimizer import Optimizer
from tqdm import tqdm
from StructuredGraphLearning.objectives import Objectives

class LearnGraphTopology:
    def __init__(self, S, is_data_matrix=False, alpha=0, maxiter=10000, abstol = 1e-6, reltol = 1e-4,
    record_objective = False, record_weights = False):
        self.S = S
        self.is_data_matrix = is_data_matrix
        self.alpha = alpha
        self.maxiter = maxiter
        self.abstol = abstol
        self.reltol = reltol
        self.record_objective = record_objective
        self.record_weights = record_weights
        self.op = Operators()
        self.obj = Objectives()
        self.optimizer = Optimizer()
        self.bic = 0

    def learn_k_component_graph(self, k=1, rho=1e-2, beta=1e4, w0='naive', fix_beta=True, beta_max = 1e6,
    lb=0, ub=1e10, eigtol = 1e-9, eps = 1e-4):
        # number of nodes
        n = self.S.shape[0]
        # find an appropriate inital guess
        if self.is_data_matrix or self.S.shape[0] != self.S.shape[1]:
            raise Exception('Not implemented yet!')
        else:
            Sinv = np.linalg.pinv(self.S)
        # if w0 is either "naive" or "qp", compute it, else return w0
        w0 = self.optimizer.w_init(w0, Sinv)
        # compute quantities on the initial guess
        Lw0 = self.op.L(w0)
        # l1-norm penalty factor
        H = self.alpha * (np.eye(n) - np.ones((n, n)))
        K = self.S + H
        U0 = self.optimizer.U_update(Lw = Lw0, k = k)
        lamda0 = self.optimizer.lamda_update(lb = lb, ub = ub, beta = beta, U = U0, Lw = Lw0, k = k)
        # save objective function value at initial guess
        if self.record_objective:
            nll0 = self.obj.negloglikelihood(Lw = Lw0, lamda = lamda0, K = K)
            fun0 = nll0 + self.obj.prior(beta = beta, Lw = Lw0, lamda = lamda0, U = U0)
            fun_seq = [fun0]
            nll_seq = [nll0]

        beta_seq = [beta]
        if self.record_weights:
            w_seq = [w0]
        time_seq = [0]
        
        start_time = time.time()
        for _ in tqdm(range(self.maxiter)):
            w = self.optimizer.w_update(w = w0, Lw = Lw0, U = U0, beta = beta, lamda = lamda0, K = K)
            Lw = self.op.L(w)
            U = self.optimizer.U_update(Lw = Lw, k = k)
            lamda = self.optimizer.lamda_update(lb = lb, ub = ub, beta = beta, U = U, Lw = Lw, k = k)
            
            # compute negloglikelihood and objective function values
            if self.record_objective:
                nll = self.obj.negloglikelihood(Lw = Lw, lamda = lamda, K = K)
                fun = nll + self.obj.prior(beta = beta, Lw = Lw, lamda = lamda, U = U)
                nll_seq.append(nll)
                fun_seq.append(fun)
            if self.record_weights:
                w_seq.append(w)
            
            # check for convergence
            werr = np.abs(w0 - w)
            has_w_converged = all(werr <= .5 * self.reltol * (w + w0)) or all(werr <= self.abstol)
            time_seq.append( time.time() - start_time )
            if not fix_beta:
                eigvals, _ = np.linalg.eigh(Lw)
            
            if not fix_beta:
                n_zero_eigenvalues = np.sum(np.abs(eigvals) < eigtol)
                if k <= n_zero_eigenvalues:
                    beta = (1 + rho) * beta
                elif k > n_zero_eigenvalues:
                    beta = beta / (1 + rho)
                if beta > beta_max:
                    beta = beta_max
                beta_seq.append(beta)
            
            if has_w_converged:
                break
            
            # update estimates
            w0 = w
            U0 = U
            lamda0 = lamda
            Lw0 = Lw
            K = self.S + H / (-Lw + eps)
        
        # compute the adjancency matrix
        Aw = self.op.A(w)
        results = {'laplacian' : Lw, 'adjacency' : Aw, 'w' : w, 'lamda' : lamda, 'U' : U,
        'elapsed_time' : time_seq, 'convergence' : has_w_converged, 'beta_seq' : beta_seq }
        if self.record_objective:
            results['obj_fun'] = fun_seq
            results['nll'] = nll_seq
            results['bic'] = 0

        if self.record_weights:
            results['w_seq'] = w_seq
        return results
    
    def learn_bipartite_graph(self, z = 0, nu = 1e4, m=7, w0='naive'):
        # number of nodes
        n = self.S.shape[0]
        # find an appropriate inital guess
        if self.is_data_matrix or self.S.shape[0] != self.S.shape[1]:
            raise Exception('Not implemented yet!')
        else:
            Sinv = np.linalg.pinv(self.S)
        # note now that S is always some sort of similarity matrix
        J = np.ones((n, n))*(1/n)
        # l1-norm penalty factor
        H = self.alpha * (2*np.eye(n) - np.ones((n, n)))
        K = self.S + H
        # if w0 is either "naive" or "qp", compute it, else return w0
        w0 = self.optimizer.w_init(w0, Sinv)
        Lips = 1 / min(np.linalg.eigvals(self.op.L(w0) + J)) 
        # compute quantities on the initial guess
        Aw0 = self.op.A(w0)
        V0 = self.optimizer.V_update(Aw0, z)
        psi0 = self.optimizer.psi_update(V0, Aw0)
        Lips_seq = [Lips]
        time_seq = [0]
        start_time = time.time()
        ll0 = self.obj.bipartite_nll(Lw = self.op.L(w0), K = K, J = J)
        fun0 = ll0 + self.obj.bipartite_prior(nu = nu, Aw = Aw0, psi = psi0, V = V0)
        fun_seq = [fun0]
        nll_seq = [ll0]
        if self.record_weights:
            w_seq = [w0]
        for _ in tqdm(range(self.maxiter)):
            # we need to make sure that the Lipschitz constant is large enough
            # in order to avoid divergence
            while(1):
                # compute the update for w
                w = self.optimizer.bipartite_w_update(w = w0, Aw = Aw0, V = V0, nu = nu, psi = psi0,
                K = K, J = J, Lips = Lips)
                # compute the objective function at the updated value of w
                fun_t = self.obj.bipartite_obj(Aw = self.op.A(w), Lw = self.op.L(w), V = V0, psi = psi0,
                K = K, J = J, nu = nu)
                # check if the previous value of the objective function is
                # smaller than the current one
                Lips_seq.append(Lips)
                if fun0 < fun_t:
                    # in case it is in fact larger, then increase Lips and recompute w
                    Lips = 2 * Lips
                else:
                    # otherwise decrease Lips and get outta here!
                    Lips = .5 * Lips
                    if Lips < 1e-12:
                        Lips = 1e-12
                    break
            Lw = self.op.L(w)
            Aw = self.op.A(w)
            V = self.optimizer.V_update(Aw = Aw, z = z)
            psi = self.optimizer.psi_update(V = V, Aw = Aw)
            # compute negloglikelihood and objective function values
            ll = self.obj.bipartite_nll(Lw = Lw, K = K, J = J)
            fun = ll + self.obj.bipartite_prior(nu = nu, Aw = Aw, psi = psi, V = V)
            # save measurements of time and objective functions
            time_seq.append(time.time() - start_time)
            nll_seq.append(ll)
            fun_seq.append(fun)
            # compute the relative error and check the tolerance on the Adjacency
            # matrix and on the objective function
            if self.record_weights:
                w_seq.append(w)
             # check for convergence
            werr = np.abs(w0 - w)
            has_w_converged = all(werr <= .5 * self.reltol * (w + w0)) or all(werr <= self.abstol)
            
            if has_w_converged:
                break
            # update estimates
            fun0 = fun
            w0 = w
            V0 = V
            psi0 = psi
            Aw0 = Aw

        results = {'laplacian' : Lw, 'adjacency' : Aw, 'w' : w, 'psi' : psi, 'V' : V,
        'elapsed_time' : time_seq, 'Lips_seq' : Lips_seq, 'convergence' : has_w_converged, 'nu' : nu }
        
        if self.record_objective:
            results['obj_fun'] = fun_seq
            results['nll'] = nll_seq
            results['bic'] = 0

        if self.record_weights:
            results['w_seq'] = w_seq

        return results



import numpy as np
import math
import time
from StructuredGraphLearning.utils import Operators
from quadprog import solve_qp

class Optimizer:
    def __init__(self):
        self.op = Operators()
        self.eps = 1e-6

    def w_init(self, w0, Sinv):
        if w0.isalpha():
            if w0 == 'qp':
                R = self.op.vecLmat(Sinv.shape[1])
                G = R.T @ R
                a = R.T @ Sinv.flatten()
                C = np.eye(R.shape[1])
                b = np.zeros(R.shape[1])
                xf, _, _, _, _, _ = solve_qp(G, a, C, b)
                w0 = xf
            elif w0 == 'naive':
                w0 = self.op.Linv(Sinv)
                w0[w0<0] = 0
        return w0

    def w_update(self, w, Lw, U, beta, lamda, K):
        c = self.op.Lstar(U @ np.diag(lamda) @ U.T - K / beta)
        grad_f = self.op.Lstar(Lw) - c
        M_grad_f = self.op.Lstar(self.op.L(grad_f))
        wT_M_grad_f = sum(w * M_grad_f)
        dwT_M_dw = sum(grad_f * M_grad_f)
        # exact line search
        t = (wT_M_grad_f - sum(c * grad_f)) / dwT_M_dw
        w_update = w - t * grad_f
        w_update[w_update < 0] = 0
        return w_update

    def bipartite_w_update(self, w, Aw, V, nu, psi, K, J, Lips):
        grad_h = 2 * w - self.op.Astar(V @ np.diag(psi) @ V.T) #+ Lstar(K) / beta
        w_update = w - (self.op.Lstar(np.linalg.inv(self.op.L(w) + J) + K) + nu * grad_h) / (2 * nu + Lips)
        w_update[w_update < 0] = 0
        return w_update

    def U_update(self, Lw, k):
        _, eigvec = np.linalg.eigh(Lw)
        p = Lw.shape[1]
        return eigvec[:, k:p]
    
    def V_update(self, Aw, z):
        p = Aw.shape[1]
        _, V = np.linalg.eigh(Aw)
        return np.hstack((V[:, 0:int(.5*(p - z))], V[:, int(.5*(p + z)):p])) 

    def psi_update(self, V, Aw, lb = float('-inf'), ub = float('inf')):
        c = np.diag(V.T @ Aw @ V)
        n = len(c)
        temp = c[int(n/2):n]
        temp = temp[::-1]
        c_tilde = .5 * (temp - c[0:int(n/2)])
        from sklearn.isotonic import IsotonicRegression
        indices = list(range(1, c_tilde.shape[0]+1))
        c_tilde = c_tilde[::-1]
        iso_reg = IsotonicRegression().fit(indices, c_tilde)
        x = iso_reg.predict(indices)
        x = np.concatenate((-x[::-1], x))
        x[x < lb] = lb
        x[x > ub] = ub
        return x

    def lamda_update(self, lb, ub, beta, U, Lw, k):
        q = Lw.shape[1] - k
        d = np.diag(U.T @ Lw @ U)
        # unconstrained solution as initial point
        lamda = 0.5 * (d + np.sqrt(d**2 + 4 / beta))
        eps = 1e-9
        condition = list((lamda[1:q-1] - lamda[0:(q-2)]) >= -eps)
        condition.append((lamda[q-1] - ub) <= eps)
        condition.append((lamda[0] - lb) >= -eps) 
        if all(condition):
            return (lamda)
        else:
            greater_ub = lamda > ub
            lesser_lb = lamda < lb
            lamda[greater_ub] = ub
            lamda[lesser_lb] = lb

        condition = list((lamda[1:q-1] - lamda[0:(q-2)]) >= -eps)
        condition.append((lamda[q-1] - ub) <= eps)
        condition.append((lamda[0] - lb) >= -eps) 
        if all(condition):
            return lamda
        else:
            print(lamda)
            raise Exception("eigenvalues are not in increasing order, consider increasing the value of beta")


from StructuredGraphLearning.LearnGraphTopology import LearnGraphTopology
from StructuredGraphLearning.utils import Operators
from StructuredGraphLearning.metrics import Metrics, ModelSelection
from StructuredGraphLearning.objectives import Objectives
from StructuredGraphLearning.optimizer import Optimizer